==> puyapy-2.1.1-py3-none-any.whl: puya/awst_build/eb/_bytes_backed.py <==
import abc
import typing
from collections.abc import Sequence

import mypy.nodes
import typing_extensions

from puya.awst.nodes import Expression, ReinterpretCast
from puya.awst_build import pytypes
from puya.awst_build.eb import _expect as expect
from puya.awst_build.eb._base import FunctionBuilder, InstanceExpressionBuilder
from puya.awst_build.eb._utils import cast_to_bytes
from puya.awst_build.eb.bytes import BytesExpressionBuilder
from puya.awst_build.eb.factories import builder_for_instance
from puya.awst_build.eb.interface import InstanceBuilder, NodeBuilder, TypeBuilder
from puya.parse import SourceLocation

_TPyType_co = typing_extensions.TypeVar(
    "_TPyType_co", bound=pytypes.PyType, default=pytypes.PyType, covariant=True
)


class BytesBackedTypeBuilder(TypeBuilder[_TPyType_co], abc.ABC):
    @typing.override
    def member_access(
        self, name: str, expr: mypy.nodes.Expression, location: SourceLocation
    ) -> NodeBuilder:
        typ = self.produces()
        match name:
            case "from_bytes":
                return _FromBytes(typ, location)
            case _:
                return super().member_access(name, expr, location)


class _FromBytes(FunctionBuilder):
    def __init__(self, result_type: pytypes.PyType, location: SourceLocation):
        super().__init__(location)
        self.result_type = result_type

    @typing.override
    def call(
        self,
        args: Sequence[NodeBuilder],
        arg_kinds: list[mypy.nodes.ArgKind],
        arg_names: list[str | None],
        location: SourceLocation,
    ) -> InstanceBuilder:
        arg = expect.exactly_one_arg_of_type_else_dummy(
            args, pytypes.BytesType, location, resolve_literal=True
        )
        result_expr = ReinterpretCast(
            expr=arg.resolve(), wtype=self.result_type.wtype, source_location=location
        )
        return builder_for_instance(self.result_type, result_expr)


class BytesBackedInstanceExpressionBuilder(InstanceExpressionBuilder[_TPyType_co], abc.ABC):
    _bytes_member: typing.ClassVar[str]

    def __init_subclass__(cls, *, bytes_member: str = "bytes", **kwargs: object):
        super().__init_subclass__(**kwargs)
        cls._bytes_member = bytes_member

    @typing.override
    def member_access(
        self, name: str, expr: mypy.nodes.Expression, location: SourceLocation
    ) -> NodeBuilder:
        if name == self._bytes_member:
            return self.bytes(location)
        else:
            return super().member_access(name, expr, location)

    def bytes(self, location: SourceLocation) -> BytesExpressionBuilder:
        return BytesExpressionBuilder(self.to_bytes(location))

    @typing.override
    def to_bytes(self, location: SourceLocation) -> Expression:
        return cast_to_bytes(self.resolve(), location)

==> puyapy-2.3.0-py3-none-any.whl: puya/awst_build/eb/_bytes_backed.py <==
import abc
import typing
from collections.abc import Sequence

import mypy.nodes
import typing_extensions

from puya.awst.nodes import Expression, ReinterpretCast
from puya.awst_build import pytypes
from puya.awst_build.eb import _expect as expect
from puya.awst_build.eb._base import FunctionBuilder, InstanceExpressionBuilder
from puya.awst_build.eb._utils import cast_to_bytes
from puya.awst_build.eb.bytes import BytesExpressionBuilder
from puya.awst_build.eb.factories import builder_for_instance
from puya.awst_build.eb.interface import InstanceBuilder, NodeBuilder, TypeBuilder
from puya.parse import SourceLocation

_TPyType_co = typing_extensions.TypeVar(
    "_TPyType_co", bound=pytypes.PyType, default=pytypes.PyType, covariant=True
)


class BytesBackedTypeBuilder(TypeBuilder[_TPyType_co], abc.ABC):
    @typing.override
    def member_access(self, name: str, location: SourceLocation) -> NodeBuilder:
        typ = self.produces()
        match name:
            case "from_bytes":
                return _FromBytes(typ, location)
            case _:
                return super().member_access(name, location)


class _FromBytes(FunctionBuilder):
    def __init__(self, result_type: pytypes.PyType, location: SourceLocation):
        super().__init__(location)
        self.result_type = result_type

    @typing.override
    def call(
        self,
        args: Sequence[NodeBuilder],
        arg_kinds: list[mypy.nodes.ArgKind],
        arg_names: list[str | None],
        location: SourceLocation,
    ) -> InstanceBuilder:
        arg = expect.exactly_one_arg_of_type_else_dummy(
            args, pytypes.BytesType, location, resolve_literal=True
        )
        result_expr = ReinterpretCast(
            expr=arg.resolve(), wtype=self.result_type.wtype, source_location=location
        )
        return builder_for_instance(self.result_type, result_expr)


class BytesBackedInstanceExpressionBuilder(InstanceExpressionBuilder[_TPyType_co], abc.ABC):
    _bytes_member: typing.ClassVar[str]

    def __init_subclass__(cls, *, bytes_member: str = "bytes", **kwargs: object):
        super().__init_subclass__(**kwargs)
        cls._bytes_member = bytes_member

    @typing.override
    def member_access(self, name: str, location: SourceLocation) -> NodeBuilder:
        if name == self._bytes_member:
            return self.bytes(location)
        else:
            return super().member_access(name, location)

    def bytes(self, location: SourceLocation) -> BytesExpressionBuilder:
        return BytesExpressionBuilder(self.to_bytes(location))

    @typing.override
    def to_bytes(self, location: SourceLocation) -> Expression:
        return cast_to_bytes(self.resolve(), location)
